/*
  ==============================================================================

    LoopbackRecording.cpp
    Created: 17 Dec 2020 10:29:16am
    Author:  JP music

  ==============================================================================
*/

#include "LoopbackRecording.h"

#include <windows.h>
#include <synchapi.h>


void data_callback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    if (pDevice == nullptr)
        return;

    CallbackData* callbackData = (CallbackData*)pDevice->pUserData;
    jassert(callbackData != nullptr);

    // Assuming we have 2 channels
    AudioDataConverters::deinterleaveSamples((const float*)pInput, callbackData->buffer.getArrayOfWritePointers(), (int)frameCount, 2);

    callbackData->fifo.addToFifo(&callbackData->buffer, frameCount);

    (void)pOutput;
}

ma_result stop__wasapi(ma_device* pDevice)
{
    jassert(pDevice != NULL);

    /*
    It's possible for the main loop to get stuck if the device is disconnected.

    In loopback mode it's possible for WaitForSingleObject() to get stuck in a deadlock when nothing is being played. When nothing
    is being played, the event is never signalled internally by WASAPI which means we will deadlock when stopping the device.
    */
    if (pDevice->type == ma_device_type_capture || pDevice->type == ma_device_type_duplex || pDevice->type == ma_device_type_loopback) {
        SetEvent((HANDLE)pDevice->wasapi.hEventCapture);
    }

    if (pDevice->type == ma_device_type_playback || pDevice->type == ma_device_type_duplex) {
        SetEvent((HANDLE)pDevice->wasapi.hEventPlayback);
    }

    return MA_SUCCESS;
}

LoopbackRecorder::LoopbackRecorder()
{

    backends[0] = ma_backend_wasapi;

    result = ma_context_init(backends, sizeof(backends) / sizeof(backends[0]), NULL, &context);
    if (result != MA_SUCCESS) {
        DBG("Failed to initialize context.");
        jassertfalse;
    }

    result = ma_context_get_devices(&context, &pPlaybackInfos, &playbackCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        DBG("Failed to retrieve list of playback devices.");
        jassertfalse;
    }

    // TODO: make sure to handle changes in devices availability (disconnected / reconnected)
    for (ma_uint32 iDevice = 0; iDevice < playbackCount; iDevice += 1) {
        deviceNames.add((String)CharPointer_UTF8(pPlaybackInfos[iDevice].name));

        if (pPlaybackInfos[iDevice].isDefault)
            defaultDeviceIndex = iDevice;
    }

    initialize(defaultDeviceIndex);
}

LoopbackRecorder::~LoopbackRecorder()
{
   releaseDevice();
    
    // juce doesn't like this when it tries to uninit its own backend stuff
    //ma_context_uninit(&context);

    timeSliceThread.removeTimeSliceClient(this);
}

void LoopbackRecorder::initialize(int deviceIndex/* = -1*/, uint32 sampleRate /*= 0*/)
{
    jassert(deviceIndex >= 0 && deviceIndex <= (int)playbackCount - 1);

    // Just in case, if device wasn't stopped before calling initialize(), calling it here.
    if (ma_device_is_started(&device))
        releaseDevice();

    bStopDevice = false;

    deviceConfig = ma_device_config_init(ma_device_type_loopback);
    deviceConfig.capture.format = ma_format_f32;
    deviceConfig.capture.pDeviceID = (deviceIndex == - 1) ? NULL : &pPlaybackInfos[deviceIndex].id;
    deviceConfig.capture.channels = 2;
    deviceConfig.sampleRate = sampleRate; // set to 0 to use device's native sampling rate
    deviceConfig.dataCallback = data_callback;
    deviceConfig.pUserData = &dataBuffer;

    result = ma_device_init(&context, &deviceConfig, &device);
    if (result != MA_SUCCESS) {
        DBG("Failed to initialize loopback device.");
        jassertfalse;
    }

    result = ma_device_start(&device);
    if (result != MA_SUCCESS) {
        ma_device_uninit(&device);
        DBG("Failed to start device.");
        jassertfalse;
    }
}

void LoopbackRecorder::releaseDevice()
{
	if (ma_device_is_started(&device)) {

		bStopDevice = true;

        // Need to signal pDevice->wasapi.hEventCapture from a different thread
        // in case Message thread gets deadlocked.
        // Here 10 is an arbitrary number, might be able to get away with shorter time.
        timeSliceThread.addTimeSliceClient(this, 10);
        timeSliceThread.startThread();

		result = ma_device_stop(&device);

        // If gotten to this point, no need to signal the event, we are out of the deadlock loop territory.
		bStopDevice = false;

		if (result != MA_SUCCESS) {
			jassertfalse;
		}
	}

    // Technically this calls "ma_device_stop()" if device is still running,
    // but doing ma_device_stop() explicitelly beforehand to catch the deadlock issue.
    ma_device_uninit(&device);
}

int LoopbackRecorder::useTimeSlice()
{
    if (bStopDevice) {
        stop__wasapi(&device);
    }
    return -1;
}
